/*  bcopy.c

    Source file for  bcopy  (copy blocks with error recovery).

    Copyright (C) 1997-1999  Richard Gooch

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  rgooch@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This programme will read blocks from a file and will write them to another
    file. Bad blocks are skipped using lseek(2) (a zero block is written to the
    output file) and are recorded in a third file for logging purposes.


    Written by      Richard Gooch   4-APR-1997

    Updated by      Richard Gooch   16-APR-1997: Packaged up.

    Last updated by Richard Gooch   19-MAY-1999: Removed declaration of
  sys_errlist.


*/
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include "version.h"

#define BLOCK_SIZE 512

static char package_id[] = PACKAGE_ID;
static char package_author[] = PACKAGE_AUTHOR;
static char package_ftp_site[] = PACKAGE_FTP_SITE;


int main (int argc, char **argv)
{
    int in_fd;
    FILE *out_fp, *log_fp;
    off_t curr_block = 0;
    ssize_t bytes_read;
    char buffer[BLOCK_SIZE];
    char *infile, *outfile, *logfile;
    static char usage_string[] = "Usage:\tcopy_blocks infile outfile logfile";

    if (argc != 4)
    {
	fprintf (stderr, "%s\n\n", usage_string);
	fprintf (stderr, "%s\n", package_id);
	fprintf (stderr, "Written by: %s\n", package_author);
	fprintf (stderr, "Available from: %s\n", package_ftp_site);
	return (1);
    }
    infile = argv[1];
    outfile = argv[2];
    logfile = argv[3];
    if ( ( in_fd = open (infile, O_RDONLY, 0) ) < 0 )
    {
	fprintf (stderr, "Error opening file: %s\t%s\n",
		 infile, sys_errlist[errno]);
	exit (2);
    }
    if ( ( out_fp = fopen (outfile, "w") ) == NULL )
    {
	fprintf (stderr, "Error opening file: %s\t%s\n",
		 outfile, sys_errlist[errno]);
	exit (2);
    }
    if ( ( log_fp = fopen (logfile, "w") ) == NULL )
    {
	fprintf (stderr, "Error opening file: %s\t%s\n",
		 logfile, sys_errlist[errno]);
	exit (2);
    }
    for (bytes_read = 1; bytes_read != 0; ++curr_block)
    {
	bytes_read = read (in_fd, buffer, BLOCK_SIZE);
	if (bytes_read == 0) break;
	if (bytes_read == -1)
	{
	    if (errno != EIO)
	    {
		fprintf (stderr, "Error reading block %ld\t%s\n",
			 curr_block, sys_errlist[errno]);
		exit (4);
	    }
	    fprintf (stderr, "Error reading block: %ld\n", curr_block);
	    fprintf (log_fp, "%ld\n", curr_block);
	    if (lseek (in_fd, BLOCK_SIZE * (curr_block + 1), SEEK_SET) == -1)
	    {
		fprintf (stderr, "Error seeking\t%s\n", sys_errlist[errno]);
		exit (3);
	    }
	    bzero (buffer, BLOCK_SIZE);
	    bytes_read = BLOCK_SIZE;
	}
	if (bytes_read != BLOCK_SIZE)
	{
	    fprintf (stderr, "Error reading block %ld\t%s\n",
		     curr_block, sys_errlist[errno]);
	    exit (4);
	}
	if (fwrite (buffer, 1, BLOCK_SIZE, out_fp) < BLOCK_SIZE)
	{
	    fprintf (stderr, "Error writing\t%s\n", sys_errlist[errno]);
	    exit (3);
	}
    }
    close (in_fd);
    fclose (out_fp);
    fclose (log_fp);
    fprintf (stderr, "Done %ld block\n", curr_block);
    exit (0);
}   /*  End Function main  */
