/*  bcopy_listed.c

    Source file for  bcopy_listed  (copy listed blocks).

    Copyright (C) 1997-1999  Richard Gooch

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  rgooch@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This programme will read blocks from a file and will write them to another
    file. A third file provides the list of blocks.


    Written by      Richard Gooch   4-APR-1997

    Updated by      Richard Gooch   16-APR-1997: Packaged up.

    Last updated by Richard Gooch   19-MAY-1999: Removed declaration of
  sys_errlist.


*/
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include "version.h"

#define BLOCK_SIZE 512
#define STRING_LENGTH 256

static char package_id[] = PACKAGE_ID;
static char package_author[] = PACKAGE_AUTHOR;
static char package_ftp_site[] = PACKAGE_FTP_SITE;


int main (int argc, char **argv)
{
    int in_fd, out_fd;
    FILE *log_in_fp, *log_out_fp;
    long slong;
    unsigned long num_blocks_recovered = 0;
    unsigned long num_blocks_failed = 0;
    off_t curr_block = 0;
    ssize_t bytes_read;
    char buffer[BLOCK_SIZE], txt[STRING_LENGTH], logfile_new[STRING_LENGTH];
    char *infile, *outfile, *logfile_old;
    static char usage_string[] =
	"Usage:\tcopy_listed_blocks infile outfile logfile";

    if (argc != 4)
    {
	fprintf (stderr, "%s\n\n", usage_string);
	fprintf (stderr, "%s\n", package_id);
	fprintf (stderr, "Written by: %s\n", package_author);
	fprintf (stderr, "Available from: %s\n", package_ftp_site);
	return (1);
    }
    infile = argv[1];
    outfile = argv[2];
    logfile_old = argv[3];
    sprintf (logfile_new, "%s.new", logfile_old);
    if ( ( in_fd = open (infile, O_RDONLY, 0) ) < 0 )
    {
	fprintf (stderr, "Error opening file: %s\t%s\n",
		 infile, sys_errlist[errno]);
	exit (2);
    }
#ifndef DEBUG
    if ( ( out_fd = open (outfile, O_RDWR, 0) ) < 0 )
    {
	fprintf (stderr, "Error opening file: %s\t%s\n",
		 outfile, sys_errlist[errno]);
	exit (2);
    }
#endif
    if ( ( log_in_fp = fopen (logfile_old, "r") ) == NULL )
    {
	fprintf (stderr, "Error opening file: %s\t%s\n",
		 logfile_old, sys_errlist[errno]);
	exit (2);
    }
    if ( ( log_out_fp = fopen (logfile_new, "w") ) == NULL )
    {
	fprintf (stderr, "Error opening file: %s\t%s\n",
		 logfile_new, sys_errlist[errno]);
	exit (2);
    }
    while (fgets (txt, STRING_LENGTH, log_in_fp) != NULL)
    {
	slong = atol (txt);
	if (slong < 0)
	{
	    fprintf (stderr, "Error reading line: %s", txt);
	    exit (3);
	}
	curr_block = slong;
	if (lseek (in_fd, BLOCK_SIZE * curr_block, SEEK_SET) == -1)
	{
	    fprintf (stderr, "Error seeking input\t%s\n", sys_errlist[errno]);
	    exit (4);
	}
	bytes_read = read (in_fd, buffer, BLOCK_SIZE);
	if (bytes_read == -1)
	{
	    if (errno != EIO)
	    {
		fprintf (stderr, "Error reading block %ld\t%s\n",
			 curr_block, sys_errlist[errno]);
		exit (5);
	    }
	    fprintf (log_out_fp, "%ld\n", curr_block);
	    ++num_blocks_failed;
	    continue;
	}
	if (bytes_read != BLOCK_SIZE)
	{
	    fprintf (stderr, "Error reading block %ld\t%s\n",
		     curr_block, sys_errlist[errno]);
	    exit (5);
	}
	/*  Hey: it worked!  */
#ifndef DEBUG
	if (lseek (out_fd, BLOCK_SIZE * curr_block, SEEK_SET) == -1)
	{
	    fprintf (stderr, "Error seeking output\t%s\n", sys_errlist[errno]);
	    exit (4);
	}
	if (write (out_fd, buffer, BLOCK_SIZE) != BLOCK_SIZE)
	{
	    fprintf (stderr, "Error writing block %ld\t%s\n",
		     curr_block, sys_errlist[errno]);
	    exit (6);
	}
#endif
	fprintf (stderr, "Recovered block: %ld\n", curr_block);
	++num_blocks_recovered;
    }
    close (in_fd);
#ifndef DEBUG
    close (out_fd);
#endif
    fclose (log_in_fp);
    fclose (log_out_fp);
#ifdef DEBUG
    fprintf (stderr, "Recovered: %lu blocks, %lu failed; debug mode\n",
	     num_blocks_recovered, num_blocks_failed);
#else
    fprintf (stderr, "Recovered: %lu blocks, %lu failed\n",
	     num_blocks_recovered, num_blocks_failed);
#endif
}   /*  End Function main  */
