/*  e2fsfind.c

    Source file for  e2fsfind  (find inodes in an ext2 filesystem).

    Copyright (C) 1997-1999  Richard Gooch

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  rgooch@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This programme will find inodes which conform to specified restrictions in
    a Linux ext2 filesystem.


    Written by      Richard Gooch   6-APR-1997

    Updated by      Richard Gooch   7-APR-1997: Added "startstring".

    Updated by      Richard Gooch   16-APR-1997: Packaged up.

    Last updated by Richard Gooch   19-MAY-1999: Removed declaration of
  sys_errlist.


*/
#include <stdio.h>
#include <math.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>
#include <sys/times.h>
#include <errno.h>
#include <utime.h>
#include <linux/fs.h>
#include <linux/ext2_fs.h>
#include "ext2fs/ext2fs.h"
#include "version.h"

#ifndef STATIC_FUNCTION
#  ifdef __cplusplus
#    define STATIC_FUNCTION( rtn, args ) extern "C" { static rtn args; }
#  else
#    ifdef c_plusplus
#      define STATIC_FUNCTION( rtn, args ) static rtn args
#    else
#      if defined(__STDC__) || defined(__stdc__)
#        define STATIC_FUNCTION( rtn, args ) static rtn args
#      else
#        define STATIC_FUNCTION( rtn, args ) static rtn()
#      endif
#    endif
#  endif
#endif

#define TRUE 1
#define FALSE 0
#define STRING_LENGTH 256

#define TYPE_ACTION  0
#define TYPE_VALUE   1
#define TYPE_STRING  2
#define TYPE_TIME    3


typedef int flag;

typedef struct _restriction_type
{
    unsigned int type;
    char name[STRING_LENGTH];
    unsigned long value;
    char string[STRING_LENGTH];
    int strlen;
    time_t timeval;
    /*TestFunc func;*/
    flag (*func) (struct _restriction_type *entry, ext2_filsys fs,
		  struct ext2_inode *inode, ino_t ino);
    struct _restriction_type *prev;
    struct _restriction_type *next;
} *restriction_type;

typedef flag (*TestFunc) (struct _restriction_type *entry, ext2_filsys fs,
			  struct ext2_inode *inode, ino_t ino);

typedef struct
{
    FILE *fp;
    char *buffer;
    int	bytes_left;
    flag failed;
} copy_info_type;

typedef struct
{
    char *buffer;
    int	bytes_left;
    restriction_type restriction;
    flag found;
} search_info_type;


/*  Private data  */
static restriction_type first_restriction = NULL;
static restriction_type last_restriction = NULL;
static FILE *logfile_fp = NULL;
static struct timeval current_time;
static copy_info_type copy_info;
static search_info_type search_info;

static char package_id[] = PACKAGE_ID;
static char package_author[] = PACKAGE_AUTHOR;
static char package_ftp_site[] = PACKAGE_FTP_SITE;


/*  Private functions  */
STATIC_FUNCTION (void process_command,
		 (ext2_filsys fs, char *device, char *line) );
STATIC_FUNCTION (void do_scan, (ext2_filsys fs, char *device, flag extract) );
STATIC_FUNCTION (flag check_inode, (ext2_filsys fs, struct ext2_inode *inode,
				    ino_t ino) );
STATIC_FUNCTION (void show_restrictions, (restriction_type list) );
STATIC_FUNCTION (void copy_inode,
		 (ext2_filsys fs, struct ext2_inode *inode, ino_t ino) );
STATIC_FUNCTION (int copy_func,
		 (ext2_filsys fs, blk_t *blocknum, int blockcount,
		  void *info) );
STATIC_FUNCTION (int search_func,
		 (ext2_filsys fs, blk_t *blocknum, int blockcount,
		  void *info) );
STATIC_FUNCTION (int search_first_func,
		 (ext2_filsys fs, blk_t *blocknum, int blockcount,
		  void *info) );
STATIC_FUNCTION (flag find_str, (char *haystack, int hay_len,
				 char *needle, int needle_len) );
STATIC_FUNCTION (void remove_restriction, (char *p) );
STATIC_FUNCTION (void show_help, () );
STATIC_FUNCTION (flag test_minsize,
		 (struct _restriction_type *entry, ext2_filsys fs,
		  struct ext2_inode *inode, ino_t ino) );
STATIC_FUNCTION (flag test_maxsize,
		 (struct _restriction_type *entry, ext2_filsys fs,
		  struct ext2_inode *inode, ino_t ino) );
STATIC_FUNCTION (flag test_startstring,
		 (struct _restriction_type *entry, ext2_filsys fs,
		  struct ext2_inode *inode, ino_t ino) );
STATIC_FUNCTION (flag test_substring,
		 (struct _restriction_type *entry, ext2_filsys fs,
		  struct ext2_inode *inode, ino_t ino) );
STATIC_FUNCTION (flag test_minage,
		 (struct _restriction_type *entry, ext2_filsys fs,
		  struct ext2_inode *inode, ino_t ino) );
STATIC_FUNCTION (flag test_maxage,
		 (struct _restriction_type *entry, ext2_filsys fs,
		  struct ext2_inode *inode, ino_t ino) );


void main (int argc, char **argv)
{
    int retval;
    ext2_filsys current_fs;
    char *device;
    char line[STRING_LENGTH];
    static char usage_string[] = "Usage:\te2fsfind device";

    if (argc != 2)
    {
	fprintf (stderr, "%s\n\n", usage_string);
	fprintf (stderr, "%s\n", package_id);
	fprintf (stderr, "Written by: %s\n", package_author);
	fprintf (stderr, "Available from: %s\n", package_ftp_site);
	exit (1);
    }
    device = argv[1];
    retval = ext2fs_open (device, 0, 0, 0, unix_io_manager, &current_fs);
    if (retval != 0)
    {
	com_err (device, retval, "while opening filesystem");
    }
    if ( ( copy_info.buffer = malloc (current_fs->blocksize) ) == NULL )
    {
	fprintf (stderr, "Error allocating copy buffer\n");
	exit (1);
    }
    if ( ( search_info.buffer = malloc (current_fs->blocksize) ) == NULL )
    {
	fprintf (stderr, "Error allocating copy buffer\n");
	exit (1);
    }
    fprintf (stderr, "e2fsfind> ");
    while (fgets (line, STRING_LENGTH, stdin) != NULL)
    {
	line[strlen (line) - 1] = '\0';
	process_command (current_fs, device, line);
	fprintf (stderr, "e2fsfind> ");
    }
    ext2fs_close (current_fs);
    if (logfile_fp != NULL) fclose (logfile_fp);
    fprintf (stderr, "done\n");
}   /*  End Function main  */

static void process_command (ext2_filsys fs, char *device, char *line)
/*  [SUMMARY] Process a command.
    <fs> The ext2 filesystem handle.
    <device> The name of the device file.
    <line> The command line to process.
*/
{
    unsigned int type = TYPE_ACTION;
    TestFunc func = (TestFunc) NULL;
    char *command, *p1;
    restriction_type new;
    extern restriction_type first_restriction, last_restriction;

    command = line;
    for (p1 = line; !isspace (*p1); ++p1);
    *p1 = '\0';
    for (++p1; isspace (*p1); ++p1);
    if (strcmp (command, "?") == 0) show_help ();
    else if (strcmp (command, "extract") == 0) do_scan (fs, device, TRUE);
    else if (strcmp (command, "help") == 0) show_help ();
    else if (strcmp (command, "list") == 0) do_scan (fs, device, FALSE);
    else if (strcmp (command, "logfile") == 0)
    {
	if (logfile_fp != NULL) fclose (logfile_fp);
	if (p1[0]!= '\0')
	{
	    if ( ( logfile_fp = fopen (p1, "w") ) == NULL )
	    {
		fprintf (stderr, "Error opening file: \"%s\"\t%s\n",
			 p1, sys_errlist[errno]);
		return;
	    }
	}
    }
    else if (strcmp (command, "minage") == 0)
    {
	func = test_minage;
	type = TYPE_TIME;
    }
    else if (strcmp (command, "minsize") == 0)
    {
	func = test_minsize;
	type = TYPE_VALUE;
    }
    else if (strcmp (command, "maxage") == 0)
    {
	func = test_maxage;
	type = TYPE_TIME;
    }
    else if (strcmp (command, "maxsize") == 0)
    {
	func = test_maxsize;
	type = TYPE_VALUE;
    }
    else if (strcmp (command, "remove") == 0) remove_restriction (p1);
    else if (strcmp (command, "show") == 0)
	show_restrictions (first_restriction);
    else if (strcmp (command, "startstring") == 0)
    {
	func = test_startstring;
	type = TYPE_STRING;
    }
    else if (strcmp (command, "substring") == 0)
    {
	func = test_substring;
	type = TYPE_STRING;
    }
    else fprintf (stderr, "Unrecognised command\n\n");
    if (type == TYPE_ACTION) return;
    if ( ( new = (restriction_type) malloc (sizeof *new) ) == NULL )
    {
	fprintf (stderr, "Error allocating memory\n");
	return;
    }
    new->type = type;
    new->func = func;
    new->next = NULL;
    strcpy (new->name, command);
    switch (type)
    {
      case TYPE_VALUE:
      case TYPE_TIME:
	new->value = strtoul (p1, NULL, 10);
	break;
      case TYPE_STRING:
	strcpy (new->string, p1);
	new->strlen = strlen (p1);
	break;
    }
    new->prev = last_restriction;
    if (first_restriction == NULL) first_restriction = new;
    if (last_restriction != NULL) last_restriction->next = new;
    last_restriction = new;
}   /*  End Function process_command  */

static void do_scan (ext2_filsys fs, char *device, flag extract)
/*  [SUMMARY] Scan all inodes for those matching restrictions.
    <fs> The ext2 filesystem handle.
    <device> The name of the device file.
    <extract> If TRUE, each inode matching the restrictions is extracted.
    [RETURNS] Nothing.
*/
{
    int retval;
    unsigned long num_found = 0;
    ino_t ino;
    ext2_inode_scan scan;
    struct ext2_inode inode;
    struct timezone tz;

    retval = ext2fs_open_inode_scan (fs, 0, &scan);
    if (retval != 0)
    {
	com_err (device, retval, "while starting inode scan");
    }
    gettimeofday (&current_time, &tz);
    while ( ( retval = ext2fs_get_next_inode (scan, &ino, &inode) ) == 0 )
    {
	if (ino == 0) break;
	if ( !LINUX_S_ISREG (inode.i_mode) ) continue;
	if (inode.i_links_count < 1) continue;
	if ( !check_inode (fs, &inode, ino) ) continue;
	fprintf (stderr, "inode: %d  links: %u\n", ino, inode.i_links_count);
	if (logfile_fp != NULL) fprintf (logfile_fp, "%d\n", ino);
	++num_found;
	if (extract) copy_inode (fs, &inode, ino);
    }
    if (retval != 0) com_err (device, retval, "while scanning inodes");
    ext2fs_close_inode_scan (scan);
    if (num_found == 1) fprintf (stderr, "1 inode matches specifications\n");
    else fprintf (stderr, "%lu inodes match specifications\n", num_found);
}   /*  End Function do_scan  */

static flag check_inode (ext2_filsys fs, struct ext2_inode *inode, ino_t ino)
/*  [SUMMARY] Check if an inode passes the tests.
    <fs> The ext2 filesystem handle.
    <inode> The inode.
    <ino> The inode number.
    [RETURNS] TRUE if the inode passes the tests, else FALSE.
*/
{
    restriction_type curr;
    extern restriction_type first_restriction;

    for (curr = first_restriction; curr != NULL; curr = curr->next)
    {
	if ( !(*curr->func) (curr, fs, inode, ino) ) return (FALSE);
    }
    return (TRUE);
}   /*  End Function check_inode  */

static void show_restrictions (restriction_type list)
/*  [SUMMARY] Show the list of restrictions.
    <list> The list.
    [RETURNS] Nothing.
*/
{
    unsigned int count;

    if (list == NULL)
    {
	fprintf (stderr, "No restrictions specified\n\n");
	return;
    }
    fprintf (stderr, "ID\t%-20sTest Value\n", "Test Name");
    for (count = 0; list != NULL; list = list->next, ++count)
    {
	fprintf (stderr, "%u\t%-20s", count, list->name);
	switch (list->type)
	{
	  case TYPE_VALUE:
	    fprintf (stderr, "%lu\n", list->value);
	    break;
	  case TYPE_STRING:
	    fprintf (stderr, "%s\n", list->string);
	    break;
	  case TYPE_TIME:
	    fprintf (stderr, "%-7lu days\n", list->value);
	    break;
	}
    }
    fprintf (stderr, "\n");
}   /*  End Function show_restrictions  */

static void copy_inode (ext2_filsys fs, struct ext2_inode *inode, ino_t ino)
/*  [SUMMARY] Copy an inode to a file.
    <fs> The ext2 filesystem handle.
    <inode> The inode.
    <ino> The inode number. The extracted file will be of the form "inode.%lu".
    [RETURNS] Nothing.
*/
{
    struct utimbuf utime_buf;
    char filename[STRING_LENGTH];

    sprintf (filename, "inode.%lu", ino);
    if ( ( copy_info.fp = fopen (filename, "w") ) == NULL )
    {
	fprintf (stderr, "Error opening file: \"%s\"\t%s\n",
		 filename, sys_errlist[errno]);
	return;
    }
    copy_info.bytes_left = inode->i_size;
    copy_info.failed = FALSE;
    ext2fs_block_iterate (fs, ino, BLOCK_FLAG_HOLE | BLOCK_FLAG_DATA_ONLY,
			  NULL, copy_func, &copy_info);
    fclose (copy_info.fp);
    if (copy_info.failed) unlink (filename);
    else
    {
	chmod (filename, inode->i_mode);
	utime_buf.actime = inode->i_atime;
	utime_buf.modtime = inode->i_mtime;
	utime (filename, &utime_buf);
    }
}   /*  End Function copy_inode  */

static int copy_func (ext2_filsys fs, blk_t *blocknum, int blockcount,
		      void *info)
/*  [SUMMARY] Copy inode callback function.
    [RETURNS] 0 on success, else BLOCK_ABORT.
*/
{
    int nbytes;
    copy_info_type *copy_info = (copy_info_type *) info;

    if (blockcount < 0) return (0);
    nbytes = (copy_info->bytes_left >
	      fs->blocksize) ? fs->blocksize : copy_info->bytes_left;
    if (*blocknum == 0)
    {
	/*  Inode has a hole: seek on output so that it has a hole as well  */
	if (fseek (copy_info->fp, nbytes, SEEK_CUR) != 0)
	{
	    copy_info->failed = TRUE;
	    return (BLOCK_ABORT);
	}
    }
    else
    {
	if (io_channel_read_blk (fs->io, *blocknum, 1, copy_info->buffer) != 0)
	    return (BLOCK_ABORT);
    }
    copy_info->bytes_left -= nbytes;
    if (fwrite (copy_info->buffer, 1, nbytes, copy_info->fp) < nbytes)
    {
	copy_info->failed = TRUE;
	return (BLOCK_ABORT);
    }
    return (0);
}   /*  End Function copy_func  */

static int search_func (ext2_filsys fs, blk_t *blocknum, int blockcount,
			void *info)
/*  [SUMMARY] Search inode callback function.
    [RETURNS] 0 on success, else BLOCK_ABORT.
*/
{
    int nbytes;
    search_info_type *search_info = (search_info_type *) info;

    if (blockcount < 0) return (0);
    if (*blocknum == 0) return (0);
    nbytes = (search_info->bytes_left >
	      fs->blocksize) ? fs->blocksize : search_info->bytes_left;
    if (io_channel_read_blk (fs->io, *blocknum, 1, search_info->buffer) != 0)
	return (BLOCK_ABORT);
    search_info->bytes_left -= nbytes;
    if ( find_str (search_info->buffer, nbytes,
		   search_info->restriction->string,
		   search_info->restriction->strlen) )
    {
	search_info->found = TRUE;
	return (BLOCK_ABORT);
    }
    return (0);
}   /*  End Function search_func  */

static int search_first_func (ext2_filsys fs, blk_t *blocknum, int blockcount,
			      void *info)
/*  [SUMMARY] Search first block in inode callback function.
    [RETURNS] 0 on success, else BLOCK_ABORT.
*/
{
    int nbytes;
    search_info_type *search_info = (search_info_type *) info;

    if (blockcount < 0) return (0);
    if (*blocknum == 0) return (BLOCK_ABORT);
    nbytes = (search_info->bytes_left >
	      fs->blocksize) ? fs->blocksize : search_info->bytes_left;
    if (nbytes < search_info->restriction->strlen) return (BLOCK_ABORT);
    if (io_channel_read_blk (fs->io, *blocknum, 1, search_info->buffer) != 0)
	return (BLOCK_ABORT);
    search_info->bytes_left -= nbytes;
    if (memcmp (search_info->buffer, search_info->restriction->string,
		search_info->restriction->strlen) == 0)
	search_info->found = TRUE;
    return (BLOCK_ABORT);
}   /*  End Function search_first_func  */

static flag find_str (char *haystack, int hay_len,
		      char *needle, int needle_len)
/*  [SUMMARY] Find if a substring exists in a larger string.
    <haystack> The string to search in.
    <hay_len> The size of the string to search in.
    <needle> The string to search for.
    <needle_len> The size of the string to search for.
    [RETURNS] TRUE if the substring exists, else FALSE.
*/
{
    flag found;
    int bytes_left, count;

    for (bytes_left = hay_len - needle_len; bytes_left > 0;
	 --bytes_left, ++haystack)
    {
	for (count = 0, found = TRUE; count < needle_len; ++count)
	{
	    if (haystack[count] != needle[count])
	    {
		found = FALSE;
		break;
	    }
	}
	if (found) return (TRUE);
    }
    return (FALSE);
}   /*  End Function find_str  */

static void remove_restriction (char *p)
/*  [SUMMARY] Remove a numbered restriction.
    <p> The string containing the restriction number.
    [RETURNS] Nothing.
*/
{
    long index, count;
    restriction_type curr;
    extern restriction_type first_restriction, last_restriction;

    if ( ( index = atol (p) ) < 0 )
    {
	fprintf (stderr, "Illegal index: \"%s\"\n\n", p);
	return;
    }
    for (count = 0, curr = first_restriction;
	 (count < index) && (curr != NULL);
	 ++count, curr = curr->next);
    if (curr == NULL)
    {
	fprintf (stderr, "No restriction with index: %ld found\n", index);
	return;
    }
    if (curr->prev == NULL) first_restriction = curr->next;
    else curr->prev->next = curr->next;
    if (curr->next == NULL) last_restriction = curr->prev;
    else curr->next->prev = curr->prev;
    free ( (char *) curr );
}   /*  End Function remove_restriction  */

static void show_help ()
/*  [SUMMARY] Show help.
    [RETURNS] Nothing.
*/
{
    fprintf (stderr, "Commands available in e2fsfind:\n\n");
    fprintf (stderr, "%-20sshow help\n", "?");
    fprintf (stderr, "%-20sextract inodes\n", "extract");
    fprintf (stderr, "%-20sshow help\n", "help");
    fprintf (stderr, "%-20slist inodes\n", "list");
    fprintf (stderr, "%-20sopen inode logfile\n", "logfile");
    fprintf (stderr, "%-20srestrict minimum age in days\n", "minage");
    fprintf (stderr, "%-20srestrict minimum size in bytes\n", "minsize");
    fprintf (stderr, "%-20srestrict maximum age in days\n", "maxage");
    fprintf (stderr, "%-20srestrict maximum size in bytes\n", "maxsize");
    fprintf (stderr, "%-20sremove numbered restriction\n", "remove");
    fprintf (stderr, "%-20sshow current restrictions\n", "show");
    fprintf (stderr, "%-20srestrict to inodes which start with string\n",
	     "startstring");
    fprintf (stderr, "%-20srestrict to inodes with substring\n", "substring");
    fprintf (stderr, "\n");
}   /*  End Function show_help  */


/*  Test functions follow  */

static flag test_minsize (struct _restriction_type *entry, ext2_filsys fs,
			  struct ext2_inode *inode, ino_t ino)
/*  [SUMMARY] Test if inode passes a specific test.
    <entry> The test parameters.
    <fs> The ext2 filesystem handle.
    <inode> The inode.
    <ino> The inode number.
    [RETURNS] TRUE if the inode passes the tests, else FALSE.
*/
{
    if (inode->i_size < entry->value) return (FALSE);
    return (TRUE);
}   /*  End Function test_minsize  */

static flag test_maxsize (struct _restriction_type *entry, ext2_filsys fs,
			  struct ext2_inode *inode, ino_t ino)
/*  [SUMMARY] Test if inode passes a specific test.
    <entry> The test parameters.
    <fs> The ext2 filesystem handle.
    <inode> The inode.
    <ino> The inode number.
    [RETURNS] TRUE if the inode passes the tests, else FALSE.
*/
{
    if (inode->i_size > entry->value) return (FALSE);
    return (TRUE);
}   /*  End Function test_maxsize  */

static flag test_startstring (struct _restriction_type *entry, ext2_filsys fs,
			      struct ext2_inode *inode, ino_t ino)
/*  [SUMMARY] Test if inode passes a specific test.
    <entry> The test parameters.
    <fs> The ext2 filesystem handle.
    <inode> The inode.
    <ino> The inode number.
    [RETURNS] TRUE if the inode passes the tests, else FALSE.
*/
{
    extern search_info_type search_info;

    search_info.bytes_left = inode->i_size;
    search_info.restriction = entry;
    search_info.found = FALSE;
    ext2fs_block_iterate (fs, ino, BLOCK_FLAG_HOLE | BLOCK_FLAG_DATA_ONLY,
			  NULL, search_first_func, &search_info);
    return (search_info.found);
}   /*  End Function test_startstring  */

static flag test_substring (struct _restriction_type *entry, ext2_filsys fs,
			    struct ext2_inode *inode, ino_t ino)
/*  [SUMMARY] Test if inode passes a specific test.
    <entry> The test parameters.
    <fs> The ext2 filesystem handle.
    <inode> The inode.
    <ino> The inode number.
    [RETURNS] TRUE if the inode passes the tests, else FALSE.
*/
{
    extern search_info_type search_info;

    search_info.bytes_left = inode->i_size;
    search_info.restriction = entry;
    search_info.found = FALSE;
    ext2fs_block_iterate (fs, ino, BLOCK_FLAG_HOLE | BLOCK_FLAG_DATA_ONLY,
			  NULL, search_func, &search_info);
    return (search_info.found);
}   /*  End Function test_substring  */

static flag test_minage (struct _restriction_type *entry, ext2_filsys fs,
			 struct ext2_inode *inode, ino_t ino)
/*  [SUMMARY] Test if inode passes a specific test.
    <entry> The test parameters.
    <fs> The ext2 filesystem handle.
    <inode> The inode.
    <ino> The inode number.
    [RETURNS] TRUE if the inode passes the tests, else FALSE.
*/
{
    if (current_time.tv_sec - inode->i_mtime < entry->value * 3600 * 24)
	return (FALSE);
    return (TRUE);
}   /*  End Function test_minage  */

static flag test_maxage (struct _restriction_type *entry, ext2_filsys fs,
			 struct ext2_inode *inode, ino_t ino)
/*  [SUMMARY] Test if inode passes a specific test.
    <entry> The test parameters.
    <fs> The ext2 filesystem handle.
    <inode> The inode.
    <ino> The inode number.
    [RETURNS] TRUE if the inode passes the tests, else FALSE.
*/
{
    if (current_time.tv_sec - inode->i_mtime > entry->value * 3600 * 24)
	return (FALSE);
    return (TRUE);
}   /*  End Function test_maxage  */
